/* $Header: /CVS/ka10/alu.c,v 1.3 1998/10/06 20:17:43 dseagrav Exp $ */

/* E10 simulator -- Arithmetic logic unit */

/* Mirian Crzig Lennox <mcl@cosmic.com> started this file. */

/* Add your name and email here when you make meaningful changes */
/* Then mail the changes to me so I know about them! */

/* This program is free software, you can redistribute it and/or
   modify it under the terms of the GNU Public License as published
   by the Free Software Foundation.  It is copyrighted by myself and
   all those who contribute code and annote their names above */

#include "flags.h"
#include "alu.h"

extern unsigned long pc_hi;		/* All functions in this file change this. */
extern unsigned long alu_lo;			/* Results go here. */
extern unsigned long alu_hi;
extern unsigned long alux_lo;			/* Extended results also go here. */
extern unsigned long alux_hi;

/* NOTE: all functions expect the high 14 bits of each long argument
   to be zero!!  */

/* Add two 36-bit words.  xl,xh is the first addend, yl,yh is the
   second addend; the sum is placed in alu_lo,alu_hi. */

void
pdp10_add(xl, xh, yl, yh)
	 unsigned long xl, xh, yl, yh;
{
	int fl;  /* flags determinant */

	/* sum right halfwords, carry to left */
	alu_lo = 0;
	alu_hi = xh + yh;
	ALU_CARRY(alu_lo, alu_hi);

	/* sum left halfwords */
	alu_lo += (xl + yl);
	
	/* construct flags determinant, a four bit quantity.
	   bits 1-2 hold the value 0, 1 or 2 if none, either or both summands
	   are negative.
	   bits 3-4 are the carry and negative bit of the sum, respectively. */
	   
	fl = ((xl & 0400000) >> 15)	+ ((yl & 0400000) >> 15) +
		((alu_lo & 01400000) >> 17);

	switch (fl) {
	case 000:
	case 005:
		/* normal result, clear all flags */
		FLAG_CLR(F_AROV | F_CRY0 | F_CRY0);
		break;
	case 001:
		/* positive overflow */
		FLAG_SET(F_AROV | F_CRY1 | F_TRAP1);
		FLAG_CLR(F_CRY0);
		break;
	case 006:
	case 013:
		/* non-overflow carry */
		FLAG_CLR(F_AROV);
		FLAG_SET(F_CRY0 | F_CRY1);
		break;
	case 012:
		/* negative overflow */
		FLAG_SET(F_AROV | F_CRY0 | F_TRAP1);
		FLAG_CLR(F_CRY1);
	default:
		/* should never get here */
	}
	ALU_NORM(alu_lo,alu_hi);
}


/* Subtract yl,yh from xl,xh.  The result is placed in alu_lo,alu_hi.
   In fact, the subtrahend is negated and added to the minuend via
   the pdp10_add() function. */

void
pdp10_sub(xl, xh, yl, yh)
	 unsigned long xl, xh, yl, yh;
{
	ALU_NEG(yl,yh);
	pdp10_add(xl, xh, yl, yh);
}


/* Multiply xl,xh by yl,yh.  The product is in alu_lo,alu_hi. */

void
pdp10_imul(xl, xh, yl, yh)
	 unsigned long xl, xh, yl, yh;
{
	alu_lo = alu_hi = 0;

	/* After each iteration of the loop, yl,yh is shifted right,
	   so it terminates when yl,yh equals 0. */

	while (yl || yh) {

		/* shift product left */

		ALU_SHL(alu_lo,alu_hi);

		/* if LSB of y pair is 1, add xl,xh to product and
		   shift y right.  If 0, just shift y. */

		if (yh & 1) {
			alu_lo += xl;
			alu_hi += xh;
			ALU_NORM(alu_lo,alu_hi);
		}
		ALU_SHR(yl,yh);
	}

	/* trap on overflow */

	if (ALU_ISOVF(alu_lo)) {
		FLAG_SET(F_AROV | F_TRAP1);
	}
	ALU_NORM(alu_lo,alu_hi);
}


/* Divide xl,xh by yl,yl.  The quotient goes to alu_lo,alu_hi, and
   the remainder goes to alux_lo,alux_hi.
   WARNING: This function is not yet finished.  */

void
pdp10_idiv(xl, xh, yl, yh)
	 unsigned long xl, xh, yl, yh;
{
	unsigned long nl, nh;   /* subtractor */
	int shft = 0;           /* divisor shift register */

	/* check for divide-by-zero condition */

	if (ALU_ISZERO(yl,yh) || (ALU_ISMIN(xl,xh) && ALU_ISMINUSONE(yl,yh))) {
		FLAG_SET(F_AROV | F_TRAP1 | F_DCK);
	}

	alu_lo = alu_hi = 0;
	nl = yl;
	nh = yh;
	ALU_NEG(nl,nh);

	/* leftshift divisor until greater than or equal to dividend */
	while (ALU_ISLT(yl,yh,xl,xh)) {
		ALU_SHL(yl,yh);
		++shft;
	}

	/* for each leftshift above, subtract and shift right */
	ALU_SHLX(nl,nh,shft);
	while (shft--) {
		ALU_SHL(alu_lo,alu_hi);
		if (ALU_ISLE(xl,xh,yl,yh)) {
			ALU_ADD(xl,xh,nl,nh);
			alu_hi |= 1;
		}
	}

	/* if anything is left in xl,xh it is the remainder. */
	alux_lo = xl;
	alux_hi = xh;
}
