/* DTE20 - Ten-Eleven Interface */
/* For E10-KL10 */

/* VARIABLES */

char dte_ibuf[1255];			/* Input buffer */
int dte_iptr;				/* Input pointer */
char dte_obuf[1255];			/* Output buffer */
int dte_optr;				/* Output pointer */
int dte_monitormode;			/* Monitor Mode */
int dte_to11nt,dte_to11et,dte_to11dn;	/* To-11 term flags */
int dte_to10nt,dte_to10et,dte_to10dn;	/* To-10 term flags */
unsigned long dte_pi_lvl;		/* PI level */
int dte_pi_zero;			/* PI0 enable */
unsigned long long dte_word;		/* DTE scratch word */
unsigned long dte_bytcnt;		/* Byte count */

/* FUNCTIONS */
unsigned long long cono_dte();
unsigned long long coni_dte();
unsigned long long datao_dte();

/* CODE */

unsigned long long cono_dte(){
	/* Switch based on bits in e_addr */
	if(e_addr&0020000LL){
		/* To-11 doorbell - Do what's in DTECMD (0451 unrelocated) */
		switch(exa_word(0451)){
			case 02400:
				/* Buffered input - Char to DTEF11 after
				   LF, ^G, or NUL.  Flushes input */
				printf("DTE20: CAN'T DEAL WITH CMD 2400\n");
				break;
			case 02000:
				/* This + Char also > TTY */
				printf("DTE20: CAN'T DEAL WITH CMD 2000\n");
				break;
			case 03400:
				/* DDT TTY READ.  Char to DTEF11. */
				/* (0450).  THEN, PI with DTEMTO set */
				dep_word(0450,dte_ibuf[dte_iptr]);
				if(dte_iptr>0){ dte_iptr--; } 
				/* Supposed to interrupt... */
				break;
			case 04000:
				/* DDT TTY WRITE. */
				/* Do I have the terminal? */
				if(consolemode==1){
					/* Yeah, just print it */
					dte_optr++;
					dte_obuf[dte_optr]=(exa_word(0350)&0177);
					/* printf("%c\n",dte_obuf[dte_optr]); */
					write(termfile,&dte_obuf[dte_optr],1);
					dte_optr--;
					break;
					}
				else{
					/* No, so alert the luser. */
					write(termfile,&termnlbel,1);
					/* Now, store the char to deal with
					   later */
					dte_obuf[dte_optr]=(char)(exa_word(0350)&0177);
					dte_optr++;
					if(dte_optr==1256){ dte_optr--; }
					break;
					}
			case 04400:
				/* MONITOR MODE ON */
				dte_monitormode=1;
				break;
			case 05000:
				/* MONITOR MODE OFF */
				dte_monitormode=0;
				break;
			case 05400:
				/* Monitor mode to DTEF11 */
				dep_word(0350,dte_monitormode);
				break;
			case 01001:
				/* Line clock on */
				printf("DTE20: CAN'T DEAL WITH CMD 1001\n");
				break;
			case 01000:
				/* Line clock off */
				printf("DTE20: CAN'T DEAL WITH CMD 1000\n");
				break;
			case 01002:
				/* Line clock wait */
				printf("DTE20: CAN'T DEAL WITH CMD 1002\n");
				break;
			case 01003:
				/* Return clock count? */
				printf("DTE20: CAN'T DEAL WITH CMD 1003\n");
				break;
			case 0400:
				/* HALT */
				printf("DTE20: HALT YOURSELF, MEATBAG\n");
				break;
			case 0401:
				/* DIAG FATAL ERROR (KILL CLOCK) */
				printf("DTE20: DIAG FATAL ERROR REPORTED BY 10\n");
				break;
			case 0402:
				/* DIAG ERROR HALT (KILL CLOCK) */
				printf("DTE20: DIAG ERROR HALT REPORTED BY 10\n");
				break;
			case 0403:
				/* END DIAG */
				printf("DTE20: 10 END DIAG\n");
				break;
			case 0404:
				/* END PASS */
				printf("DTE20: 10 END PASS\n");
				break;
			case 0405:
				/* GET CLOCK DEFAULT WORD */
				printf("DTE20: CAN'T DEAL WITH CMD 405\n");
				break;
			case 0406:
				/* DIAMON FILE SELECT */
				printf("DTE20: CAN'T DEAL WITH CMD 406\n");
				break;
			case 0407:
				/* DIAMON FILE READ */
				printf("DTE20: CAN'T DEAL WITH CMD 407\n");
				break;
			case 0410:
				/* KLDCP COMMAND */
				printf("DTE20: CAN'T DEAL WITH CMD 410\n");
				break;
			case 01400:
				/* LOAD SIMULATED SWITCHES */
				dep_word(0340,0);
				break;
			default:
				/* TYPE CRAP OUT */
				/* Do I have the terminal? */
				if(consolemode==1){
					/* Yeah, just print it */
					dte_optr++;
					dte_obuf[dte_optr]=(exa_word(0451)&0177);
					/* printf("%c\n",dte_obuf[dte_optr]); */
					write(termfile,&dte_obuf[dte_optr],1);
					dte_optr--;
					break;
					}
				else{
					/* No, so alert the luser. */
					write(termfile,&termnlbel,1);
					/* Now, store the char to deal with
					   later */
					dte_obuf[dte_optr]=(char)(exa_word(0451)&0177);
					dte_optr++;
					if(dte_optr==1256){ dte_optr--; }
					break;
					}
				}
				/* Switch done */
				/* Set command done */
				dep_word(0444,0777777777777LL);
		}
	if(e_addr&0010000LL){
		printf("DTE20: CR11B set on CONO DTE, ignoring\n");
		}
	if(e_addr&0004000LL){
		printf("DTE20: ST11B set, why's the 10 trying to reboot me?\n");
		}
	if(e_addr&0001000LL){
		printf("DTE20: CL11PI, but 11 never asked 10!\n");
		}
	if(e_addr&0000100LL){
		/* CLTO11 */
		dte_to11nt=0; dte_to11et=0;
		}
	if(e_addr&0000400LL){
		/* CLTO10 */
		dte_to10nt=0; dte_to10et=0;
		}
	if(e_addr&0000200LL){
		/* PILDEN */
		dte_pi_lvl=e_addr&07;
		}
	if(e_addr&0000100LL){
		/* PI0ENB */
		dte_pi_zero=1;
		}
return(0);
}

unsigned long long coni_dte(){
	dte_word=0;
	if(dte_pi_zero){ dte_word=dte_word|010; }
	if(dte_to10et){ dte_word=dte_word|0400; }
	if(dte_to11dn){ dte_word=dte_word|0100; }
	if(dte_to10dn){ dte_word=dte_word|040; }
	if(dte_to10et){ dte_word=dte_word|020; }
	dte_word=dte_word|dte_pi_lvl;
	do_memory_write(e_addr,dte_word);
		
return(0);
}

unsigned long long datao_dte(){
	dte_word=do_memory_read(e_addr,0);
	/* What's TO10IB? */
	dte_bytcnt=dte_word&07777;


return(0);
}
