/* PRIORITY INTERRUPT SYSTEM for E10-KL10 */

/* VARIABLES */
int pi_req[07];					/* PI lines 0-7 */
int pi_prog_req[07];				/* Software PI lines */
int pi_ena[07];					/* PI enabled lines */
int pi_pdn[07];					/* PI physical dev # */
unsigned long long pi_fcn[07];			/* PI function word */
int pi_system_active;				/* PI system active */
int pi_cycle;					/* PI cycle flag */
unsigned long long pi_scratch;			/* PI scratch word */
	/* FROM THE PAGER */
extern unsigned long pg_ept_page;

	/* MICROSTATE SAVE AREA */
unsigned long long pi_us_pc;			/* Saved PC */
unsigned long pi_us_cpu_section;		/* Section */
unsigned long long pi_us_vma;			/* Saved VMA */
unsigned long long pi_us_pma;
unsigned long long pi_us_ar,pi_us_arx;
unsigned long long pi_us_br,pi_us_brx;
long pi_us_sctr;
unsigned long pi_us_ma,pi_us_ir;
unsigned long pi_us_ac;
unsigned pi_us_opc;
unsigned long pi_us_idbit,pi_us_id_ac,pi_us_inst_addr;
unsigned long long pi_us_jrstf_ptr;
unsigned long long pi_us_e_addr;
unsigned long pi_us_micropc;
unsigned long long pi_us_llbuffer;
unsigned long long pi_us_bptr_size,pi_us_bptr_pos;
unsigned long long pi_us_bptr_type,pi_us_bptr_addr;
unsigned long long pi_us_bptr_ac,pi_us_bptr_loc,pi_us_bptr_ind;
int pi_us_holding_pi;

/* NOTES:

PI CYCLE:
1.	Device puts a 1 in pi_req, on it's level, it's pdn in pi_pdn[lvl],
	and a function word in pi_fcn[lvl]. (Only if it's PI level isn't 
	already high!)
2.	If the PI sys is on and the level is enabled, the processor
	saves the microstate and goes to PI service mode.

PDNs:
	Interval Counter					101
	CPU errors, program req's, other internals:		102
	Channels 0-7						0-7
	DTE20s 0-3						10-13
	DIA20 (IO BUS DEVICES)					17

*/

/* FUNCTIONS */

unsigned long long cono_pi();
unsigned long long coni_pi();
unsigned long long pi_check();
void pi_do_microsave();
void pi_undo_microsave();

/* CODE */

unsigned long long cono_pi(){
/* LOAD PI STATS FROM E */

if(e_addr&0020000LL){
	/* Drop program requests on selected lines. */
	if(e_addr&0100){ pi_prog_req[1]=0; }
	if(e_addr&040){ pi_prog_req[2]=0; }
	if(e_addr&020){ pi_prog_req[3]=0; }
	if(e_addr&010){ pi_prog_req[4]=0; }
	if(e_addr&04){ pi_prog_req[5]=0; }
	if(e_addr&02){ pi_prog_req[6]=0; }
	if(e_addr&01){ pi_prog_req[7]=0; }
	}
if(e_addr&0010000LL){
	/* Clear PI system */
	printf("%%PI: PI SYSTEM CLEAR\n");
	/* Disable PI system */
	pi_system_active=0;
	/* Turn off all levels, drop all requests */
	pi_scratch=0;
	while(pi_scratch<=7){
		pi_req[pi_scratch]=0;
		pi_prog_req[pi_scratch]=0;
		pi_ena[pi_scratch]=0;
		pi_pdn[pi_scratch]=0;
		pi_fcn[pi_scratch]=0;
		pi_scratch++;
		}
	}
	if(e_addr&0004000LL){
		/* INITIATE INTERRUPTS ON */
		/* Force the processor to initiate interrupts
		   on the selected levels.  The intr made will be
		   recognised even with the PI level off.
		   Also, these interrupts are not cleared on completion,
		   and will remain until forcibly dropped. */
		if(e_addr&0100LL){ pi_req[1]=2; }
		if(e_addr&040LL){ pi_req[2]=2; }
		if(e_addr&020LL){ pi_req[3]=2; }
		if(e_addr&010LL){ pi_req[4]=2; }
		if(e_addr&04LL){ pi_req[5]=2; }
		if(e_addr&02LL){ pi_req[6]=2; }
		if(e_addr&01LL){ pi_req[7]=2; }
		}
	if(e_addr&0002000LL){
		/* Turn on selected levels */
		if(e_addr&0100LL){ pi_ena[1]=1; }
		if(e_addr&040LL){ pi_ena[2]=1; }
		if(e_addr&020LL){ pi_ena[3]=1; }
		if(e_addr&010LL){ pi_ena[4]=1; }
		if(e_addr&04LL){ pi_ena[5]=1; }
		if(e_addr&02LL){ pi_ena[6]=1; }
		if(e_addr&01LL){ pi_ena[7]=1; }
		}
	if(e_addr&0001000LL){
		/* Turn off selected levels */
		if(e_addr&0100LL){ pi_ena[1]=0; }
		if(e_addr&040LL){ pi_ena[2]=0; }
		if(e_addr&020LL){ pi_ena[3]=0; }
		if(e_addr&010LL){ pi_ena[4]=0; }
		if(e_addr&04LL){ pi_ena[5]=0; }
		if(e_addr&02LL){ pi_ena[6]=0; }
		if(e_addr&01LL){ pi_ena[7]=0; }
		}
	if(e_addr&0000400LL){
		/* PI SYSTEM OFF */
		pi_system_active=0;
		printf("%%PISYS: PI SYSTEM DISABLED\n");
		}
	if(e_addr&0000200LL){
		/* PI SYSTEM ON */
		pi_system_active=1;
		printf("%%PISYS: PI SYSTEM ENABLED\n");
		}

return(0);
}

unsigned long long coni_pi(){

pi_scratch=0;
if(pi_prog_req[1]==1){ pi_scratch=pi_scratch|0000100000000LL; }
if(pi_prog_req[2]==1){ pi_scratch=pi_scratch|0000040000000LL; }
if(pi_prog_req[3]==1){ pi_scratch=pi_scratch|0000020000000LL; }
if(pi_prog_req[4]==1){ pi_scratch=pi_scratch|0000010000000LL; }
if(pi_prog_req[5]==1){ pi_scratch=pi_scratch|0000004000000LL; }
if(pi_prog_req[6]==1){ pi_scratch=pi_scratch|0000002000000LL; }
if(pi_prog_req[7]==1){ pi_scratch=pi_scratch|0000001000000LL; }
/* FORCE PARITY BITS IGNORED FOR NOW */
if(pi_req[1]!=0){ pi_scratch=pi_scratch|0000000040000LL; }
if(pi_req[2]!=0){ pi_scratch=pi_scratch|0000000020000LL; }
if(pi_req[3]!=0){ pi_scratch=pi_scratch|0000000010000LL; }
if(pi_req[4]!=0){ pi_scratch=pi_scratch|0000000004000LL; }
if(pi_req[5]!=0){ pi_scratch=pi_scratch|0000000002000LL; }
if(pi_req[6]!=0){ pi_scratch=pi_scratch|0000000001000LL; }
if(pi_req[7]!=0){ pi_scratch=pi_scratch|0000000000400LL; }
if(pi_system_active==1){ pi_scratch=pi_scratch|0000000000200LL; }
if(pi_ena[1]==1){ pi_scratch=pi_scratch|0000000000100LL; }
if(pi_ena[2]==1){ pi_scratch=pi_scratch|0000000000400LL; }
if(pi_ena[3]==1){ pi_scratch=pi_scratch|0000000000020LL; }
if(pi_ena[4]==1){ pi_scratch=pi_scratch|0000000000010LL; }
if(pi_ena[5]==1){ pi_scratch=pi_scratch|0000000000004LL; }
if(pi_ena[6]==1){ pi_scratch=pi_scratch|0000000000002LL; }
if(pi_ena[7]==1){ pi_scratch=pi_scratch|0000000000001LL; }
do_memory_write(e_addr,pi_scratch);

return(0);
}

unsigned long long pi_check(){
/* This is run before every microcycle, and is the main PI system loop */
if(pi_system_active==0){ return(0); }		/* PI sys has to be on */
if(pi_cycle){ return(0); }			/* Only run when not PI cycle */
pi_scratch=0;

while(pi_scratch<=7){
	if((pi_req[pi_scratch]==1&&pi_ena[pi_scratch]==1)||pi_req[pi_scratch]==2){
		/* Hardware interrupt needed */
		/* What kind of interrupt? */
		/* Not null word, deal with it */
		switch((pi_fcn[pi_scratch]&0070000000000LL)>>30){
		case 00:
			/* Internal device or zero word */
			/* For the interval counter, do a vector intr. */
			/* Else, perform a standard interrupt. */
			if(pi_pdn[pi_scratch]==101){
				/* Interval counter. */
				goto do_vector_interrupt;
				}
			else{
				/* Other. */
				goto do_std_interrupt;
				}
			break;
		case 01:
			/* Standard interrupt. */
			do_std_interrupt:
			pi_do_microsave();	/* SAVE CPU STATE */
			/* Run the instruction in 040+2N of the exec proc table */
			vma=pg_ept_page<<9;	/* VMA = base of EPT */
			vma=vma+(040+(pi_scratch*2));
			/* Load the instruction */
			arx=exa_word(vma);
			brx=arx;
		        /* Build instruction-based toys */
		        ac=(arx&(0000740000000LL))>>23;
		        id_bit=(arx&(0000020000000LL))>>22;
		        id_ac=(arx&(0000017000000LL))>>18;
			inst_addr=arx&0777777;
        		micropc=2;		/* Dispatch to XCT entry */
			pi_cycle=1;
			/* Kernel mode... */
			if(pc&F_USER){ pc=pc^F_USER; }
			pc=pc|F_PUBL;
			/* Okay, I think that's it... */
			break;

		case 02:
			/* Vector interrupt. */
do_vector_interrupt:
			if(pi_pdn[pi_scratch]==0101){
				/* Interval Counter - Execute instr at
				   EPT+514 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in 040+2N of the exec proc table */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+0514;
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
		        	/* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
			        id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
	        		micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]==010){
				/* DTE20 #0 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in 040+2N of the exec proc table */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+(0142);
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]==011){
				/* DTE20 #1 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in 040+2N of the exec proc table */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+(0152);
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]==012){
				/* DTE20 #2 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in 040+2N of the exec proc table */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+(0162);
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]==013){
				/* DTE20 #3 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in 040+2N of the exec proc table */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+(0172);
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]<=07){
				/* A CHANNEL */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Run the instruction in ept+27:35 */
				vma=pg_ept_page<<9;	/* VMA = base of EPT */
				vma=vma+(pi_fcn[pi_scratch]&0777);
				/* Load the instruction */
				arx=exa_word(vma);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Kernel mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				pc=pc|F_PUBL;
				/* Okay, I think that's it... */
				break;
				}
			if(pi_pdn[pi_scratch]==017){
				/* DIA20 */
				/* Exec instr at exec vloc 13:35 */
				pi_do_microsave();	/* SAVE CPU STATE */
				/* Exec mode... */
				if(pc&F_USER){ pc=pc^F_USER; }
				if(pc&F_PUBL){ pc=pc^F_PUBL; }
				vma=(pi_fcn[pi_scratch]&037777777LL);
				/* Load the instruction */
				arx=do_memory_read(vma,1);
				brx=arx;
			        /* Build instruction-based toys */
			        ac=(arx&(0000740000000LL))>>23;
			        id_bit=(arx&(0000020000000LL))>>22;
		        	id_ac=(arx&(0000017000000LL))>>18;
				inst_addr=arx&0777777;
        			micropc=2;		/* Dispatch to XCT entry */
				pi_cycle=1;
				/* Okay, I think that's it... */
				break;
				}
				/* End of IFs */
				break;
			case 03:
				/* Increment or DEC-rement (That's bad ^_^) */
				if(pi_fcn[pi_scratch]&0004000000000LL){
					/* Increment exec virt loc */
					pi_do_microsave();
					vma=(pi_fcn[pi_scratch]&037777777LL);
					/* Exec mode... */
					if(pc&F_USER){ pc=pc^F_USER; }
					if(pc&F_PUBL){ pc=pc^F_PUBL; }
					/* Increment */
					do_memory_write(vma,(do_memory_read(vma,0)+1));
					/* Done */
					pi_undo_microsave();
					}
				else{
					/* Decrement exec virt loc */
					pi_do_microsave();
					vma=(pi_fcn[pi_scratch]&037777777LL);
					/* Exec mode... */
					if(pc&F_USER){ pc=pc^F_USER; }
					if(pc&F_PUBL){ pc=pc^F_PUBL; }
					/* Decrement */
					do_memory_write(vma,(do_memory_read(vma,0)-1));
					/* Done */
					pi_undo_microsave();
					}
				break;
			case 04:
				/* DTE20 EXAMINE */
			case 05: 
				/* DTE20 DEPOSIT */
			case 06:
				/* DTE20 BYTE TRANSFER */
				printf("%%PISYS: DTE20 EXAMINE,DEPOSIT,BYTE XFER AREN'T IMPLEMENTED, GOMEN!\n");
				pi_req[pi_scratch]=0; pi_pdn[pi_scratch]=0; pi_fcn[pi_scratch]=0;
				break;

			case 07:
				/* RESERVED */
				printf("%%PISYS: TSUKI NO HIKARI WA AI NO MESSAGE\n");
				pi_req[pi_scratch]=0; pi_pdn[pi_scratch]=0; pi_fcn[pi_scratch]=0;
				break;
			
			/* END OF SWITCH */
			}

		}
	/* Not. */
	pi_scratch++;
	}


return(0);
}

void pi_do_microsave(){
pi_us_pc=pc;
pi_us_cpu_section=cpu_section;
pi_us_vma=vma;
pi_us_pma=pma;
pi_us_ar=ar; pi_us_arx=arx;
pi_us_br=br; pi_us_brx=brx;
pi_us_sctr=sctr;
pi_us_ma=ma; pi_us_ir=ir;
pi_us_ac=ac;
pi_us_opc=opc;
pi_us_idbit=id_bit; pi_us_id_ac=id_ac; pi_us_inst_addr=inst_addr;
pi_us_jrstf_ptr=jrstf_ptr;
pi_us_e_addr=e_addr;
pi_us_micropc=micropc;
pi_us_llbuffer=ll_buffer;
pi_us_bptr_size=bptr_size; pi_us_bptr_pos=bptr_pos;
pi_us_bptr_addr=bptr_addr; pi_us_bptr_type=bptr_type;
pi_us_bptr_loc=bptr_loc; pi_us_bptr_ac=bptr_ac;
pi_us_bptr_ind=bptr_ind;

pi_us_holding_pi=pi_scratch;

}

void pi_undo_microsave(){
pc=pi_us_pc;
cpu_section=pi_us_cpu_section;
vma=pi_us_vma;
pma=pi_us_pma;
ar=pi_us_ar; arx=pi_us_arx;
br=pi_us_br; brx=pi_us_brx;
sctr=pi_us_sctr;
ma=pi_us_ma; ir=pi_us_ir;
ac=pi_us_ac;
opc=pi_us_opc;
id_bit=pi_us_idbit; id_ac=pi_us_id_ac; inst_addr=pi_us_inst_addr;
jrstf_ptr=pi_us_jrstf_ptr;
e_addr=pi_us_e_addr;
micropc=pi_us_micropc;
ll_buffer=pi_us_llbuffer;
bptr_size=pi_us_bptr_size; bptr_pos=pi_us_bptr_pos;
bptr_type=pi_us_bptr_type; bptr_addr=pi_us_bptr_addr;
bptr_loc=pi_us_bptr_loc; bptr_ac=pi_us_bptr_ac;
bptr_ind=pi_us_bptr_ind; 
}

